#!/usr/bin/env ruby

# generate bridge support metadata files for Cocoa & its dependencies.
TIGER_OR_LOWER = `uname -r`.strip.to_f < 9.0
system('mkdir -p bridge-support')
[['/System/Library/Frameworks/CoreFoundation.framework', nil],
 ['/System/Library/Frameworks/Foundation.framework', nil],
 ['/System/Library/Frameworks/AppKit.framework', nil],
 ['/System/Library/Frameworks/WebKit.framework', nil],
 ['/System/Library/Frameworks/ApplicationServices.framework/Frameworks/CoreGraphics.framework', '-c "-framework ApplicationServices" -c -F/System/Library/Frameworks/ApplicationServices.framework/Frameworks'],
 ['/System/Library/Frameworks/Quartz.framework/Frameworks/PDFKit.framework', nil],
 ['/System/Library/Frameworks/QuartzCore.framework', nil],
 ['/System/Library/Frameworks/OpenGL.framework', nil],
 ['/System/Library/Frameworks/QTKit.framework', TIGER_OR_LOWER ? '' : '-c -DQT_BUILDING_ON_LEOPARD_OR_LATER -c "-framework QTKit"'],
].each do |path, special_flags|
  framework = File.basename(path, '.framework')
  out = "bridge-support/#{framework}.xml"
  if !File.exists?(out) or File.size(out) == 0 or File.mtime('gen_bridge_metadata.rb') > File.mtime(out)
    generator = './gen_bridge_metadata.rb'
    generator << " #{special_flags}" if special_flags
    generator << " -f #{path}"
    exceptions = "bridge-support-exceptions/#{framework}.xml"
    generator << " -e #{exceptions}" if File.exists?(exceptions)
    generator << " -o #{out}"
    $stderr.puts "create #{out} ..."
    system(generator)
  end
  # Uncomment this to launch the verification tool on each metadata file.
  # Warning: this can take some time, and there are several false positives.
  #$stderr.puts "verify #{out} ..."
  #system("ruby tool/verify_bridge_metadata.rb #{out} #{File.join(path, 'Headers')}")
end
