#import <Cocoa/Cocoa.h>
#import <objc/objc-runtime.h>

@interface NSObject(subclasses)
+ (NSArray*) subclasses;
@end

@implementation NSObject(subclasses)
+ (NSArray*) subclasses {
    NSMutableArray		*subclasses;
    struct objc_class	*superClass;
    Class				*classes = NULL;
    Class				*current;
	const Class			thisClass = [self class];
    int					count, index;

	subclasses = [[NSMutableArray alloc] initWithCapacity:12];

	count = objc_getClassList(NULL, 0);
		
	if (count) {
		classes = malloc(sizeof(Class) * count);
		NSAssert (classes != NULL, @"Memory Allocation Failed in [NSObject +subclasses].");

		(void) objc_getClassList(classes, count);
	}

	if (classes) {
		current = classes;

		for (index = 0; index < count; ++index, ++current) {
			superClass = (*current)->super_class;
						
			if (superClass == thisClass) {
				[subclasses addObject:*current];
			}
		}

		free(classes);
	}

    return subclasses;
}
@end

BOOL
classWorks(Class class) {
	NSCharacterSet *charSet = [NSCharacterSet alphanumericCharacterSet];
	NSString *className = [class className];
	int i = 0;

	for(i=0; i < [className length]; i++) 
		if(![charSet characterIsMember:[className characterAtIndex:i]]) return NO;
	
	if([[[NSBundle bundleForClass:class] bundlePath] hasPrefix:@"/System/Library/PrivateFrameworks"]) return NO;
	if([className hasPrefix:@"CI"] || [className hasPrefix:@"PDF"]) return NO;
	
	return YES;
}

void
appendClassDef(Class class, NSMutableString *output) {
	Class superClass = class;
	
	if(!classWorks(class)) return;
	while(superClass = [superClass superclass]) if(classWorks(superClass)) break;
	
	if(superClass) [output appendFormat:@"%@ : %@ {\n", [class className], [superClass className]];
	else [output appendFormat:@"%@ : CCTrampoline {\n", [class className]];
	[output appendFormat:@"\tclassName {^\"%@\"}\n", [class className]];
	[output appendFormat:@"}\n"];
}

int
classDefsForClass(Class class, NSMutableString *output) {
	unsigned int numClasses = 0;
	NSMutableArray *classes;
	NSEnumerator *enumerator;
	id current;
	
	classes = (NSMutableArray *)[class subclasses];
	enumerator = [classes objectEnumerator];
	
	while(current = [enumerator nextObject]) {
		appendClassDef(current, output);
		numClasses += classDefsForClass(current, output) + 1;
	}
	
	return numClasses;
}

int main (int argc, const char * argv[]) {
	NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];
	NSString *path;
	unsigned int numClasses = 0;
	
	if(argc != 2) {
		printf("Usage: %s outputFile\n", argv[0]);
		return 1;
	}
	else {
		path = [NSString stringWithUTF8String:argv[1]];
	}
	
	NSMutableString *output = [[NSMutableString alloc] init];
	
	[output appendString:@"/* This file is autogenerated by SCClassBuilder. These class definitions should eventually be replaced by something that dynamically adds them to the SC runtime from the Obj-C runtime, but AFAIK that isn't possible right now. The className method is necessary to support calling super from within an Obj-C subclass. The className method should only be implemented for Obj-C classes, not for SC classes that will be called from Obj-C. */\n\n"];

	appendClassDef([NSObject class], output);
	numClasses = classDefsForClass([NSObject class], output);
	
	[output writeToFile:path atomically:NO encoding:NSASCIIStringEncoding error:nil];

	NSLog(@"Wrote %d classes", numClasses);

	[pool release];
	return 0;
}
