#import <Cocoa/Cocoa.h>
#import <objc/objc.h>
#import <objc/objc-runtime.h>
#import "BetterInvocation.h"
#import "next_mapping.h"
#import "ffi_utils.h"
#import "ffi.h"

#define NSLog(format, ...)

#if defined(__ppc__)
	#define SMALL_STRUCT_LIMIT 4
#elif defined(__i386__) 
	#define SMALL_STRUCT_LIMIT 8
#else
	#error "Unsupported platform"
#endif

@implementation BetterInvocation
- (void)setSendsToSuper:(BOOL)flag {
	sendsToSuper = flag;
}

- (BOOL)sendsToSuper {
	return sendsToSuper;
}

- (void)invoke {
	if(sendsToSuper) [self superInvoke];
	else [super invoke];
}

- (void)superInvoke {
	NSMethodSignature *methodSignature = [self methodSignature];
	int numArgs = [methodSignature numberOfArguments], retSize = [methodSignature methodReturnLength];
	const char *cRetType = [methodSignature methodReturnType];
	ffi_type **argTypes, *retType;
	ffi_cif cif;
	void *retValue = NULL, **argValues;
	int curArg = 0, j = 0;;
	BOOL useStret;
	
	useStret = NO;
	if (*cRetType == _C_STRUCT_B && 
		(retSize > SMALL_STRUCT_LIMIT
#ifdef __i386__
		 /* darwin/x86 ABI is slightly odd ;-) */
		 || (retSize != 1 
			&& retSize != 2 
			&& retSize != 4 
			&& retSize != 8)
#endif
		)) {
			useStret = YES;
	}

	/* allocate an extra arg for stretAddr when calling objc_msgSendSuper_stret */
	argTypes = (ffi_type **)malloc(sizeof(ffi_type *) * numArgs);
	argValues = (void **)malloc(sizeof(void *) * numArgs);
	argValues[numArgs-1] = NULL;
	argTypes[numArgs-1] = NULL;

	struct objc_super superContx = superContext([self target]);
	struct objc_super *superContxPtr = &superContx;
	
	if(retSize > 0) retValue = malloc(retSize);
	
	retType = signature_to_ffi_type(cRetType);
	
	argTypes[curArg] = signature_to_ffi_type(@encode(struct objc_super *));
	argValues[curArg] = (struct objc_super *)&superContxPtr;
	curArg++;
	
	for(j=1; j < numArgs; j++, curArg++) {
		const char *type = [methodSignature getArgumentTypeAtIndex:j];
		argTypes[curArg] = signature_to_ffi_type(type);
		argValues[curArg] = malloc(objc_sizeof_type(type));
		[self getArgument:argValues[curArg] atIndex:j];
	}
	
	NSLog(@"calling super with ret type: %s, %d", [methodSignature methodReturnType], retSize);
	
	if (ffi_prep_cif(&cif, FFI_DEFAULT_ABI, numArgs, retType, argTypes) != FFI_OK) {
		NSLog(@"Can't prepare CIF");
		return;
	}
	
	if(useStret)
		ffi_call(&cif, FFI_FN(objc_msgSendSuper_stret), retValue, argValues);
	else
		ffi_call(&cif, FFI_FN(objc_msgSendSuper), retValue, argValues);
	
	if(retSize > 0) [self setReturnValue:retValue];
}
@end