#import <Cocoa/Cocoa.h>
#import "CCArray.h"
#import "CCBridge.h"
#import "PyrObject.h"
#import "PyrSlot.h"
#import "SC_Constants.h"
#import "PyrPrimitive.h"
#import "PyrKernel.h"
#import "GC.h"

#define NSLog(format, ...)

int prArrayAdd(struct VMGlobals *g, int numArgsPushed);
int prArrayInsert(struct VMGlobals *g, int numArgsPushed);
int prArrayPop(struct VMGlobals *g, int numArgsPushed);
int basicRemoveAt(VMGlobals *g, int numArgsPushed);

@implementation CCArray
+ (CCArray *)newWithSCArray:(PyrObject *)aValue {
	CCArray *res = [[CCArray alloc] initWithSCArray:aValue];
	[res autorelease];
	return res;
}

- (CCArray *)initWithSCArray:(PyrObject *)aValue {
	self = [super init];
	if(!self) return nil;
	scArray = aValue;
	scList = nil;
	return self;
}

+ (CCArray *)newWithSCList:(PyrObject *)aValue {
	CCArray *res = [[CCArray alloc] initWithSCList:aValue];
	[res autorelease];
	return res;
}

- (CCArray *)initWithSCList:(PyrObject *)aValue {
	self = [super init];
	if(!self) return nil;
	scArray = aValue->slots[0].uo;
	scList = aValue;
	return self;
}

- (PyrObject *)pyrObject {
	return scArray;
}

- (int)count {
	return scArray->size;
}

- (id)objectAtIndex:(int)idx {
	PyrSlot val;
	id obj;
	getIndexedSlot(scArray, &val, idx);
	sc_to_objc(@encode(id), &val, &obj);
	
	if(!obj) obj = [NSNull null];
	
	return obj;
}

- (void)replaceObjectAtIndex:(int)idx withObject:(id)newValue {
	VMGlobals *g = gMainVMGlobals;
	PyrSlot val;
	
	objc_to_sc(@encode(id), &newValue, &val, YES);
	putIndexedSlot(gMainVMGlobals, scArray, &val, idx);
	g->gc->GCWrite(scArray, &val);
}

- (void)getObjects:(id*)buffer inRange:(NSRange)range {
	unsigned int i;

	for (i = 0; i < range.length; i++) {
		buffer[i] = [self objectAtIndex:i+range.location];
	}
}

/* 
 * To avoid calling back on SC (which would result in the user having to call .defer
 * much more often), the following methods interact directly with the C array 
 * primitives. This is also reduces code duplication. The downside is that we only
 * support List and ArrayedCollection subclasses; not any SequenceableCollection
 * subclass. There probably should be a check to see that there is size left on the
 * stack before using it.
 */

- (void)addObject:(id)anObject {
	VMGlobals *g = gMainVMGlobals;
	PyrSlot val, *oldStack;

	NSLog(@"in addObject with %@", anObject);

	if((scArray->size >= MAXINDEXSIZE(scArray)) && scList == nil)
		@throw [NSException exceptionWithName:NSRangeException reason:@"No size left in (non-mutable) ArrayedCollection" userInfo:nil];

	oldStack = g->sp;
	
	objc_to_sc(@encode(id), &anObject, &val, YES);
	++g->sp; SetObject((PyrSlot *)g->sp, scArray);
	++g->sp; g->sp->ucopy = val.ucopy;
	prArrayAdd(g, 2);
	
	scArray = (g->sp-1)->uo;
	
	if(scList) {
		scList->slots[0].uo = scArray;
		g->gc->GCWrite(scList, scArray);
	}
	
	[anObject retain];
	
	g->sp = oldStack;
}

- (void)insertObject:(id)anObject atIndex:(unsigned)idx {
	VMGlobals *g = gMainVMGlobals;
	PyrSlot val, *oldStack;
	
	NSLog(@"in insertObject with %@", anObject);
	
	if((scArray->size >= MAXINDEXSIZE(scArray)) && scList == nil)
		@throw [NSException exceptionWithName:NSRangeException reason:@"No size left in (non-mutable) ArrayedCollection" userInfo:nil];

	oldStack = g->sp;
	
	objc_to_sc(@encode(id), &anObject, &val, YES);
	++g->sp; SetObject((PyrSlot *)g->sp, scArray);
	++g->sp; SetInt((PyrSlot *)g->sp, (int)idx);
	++g->sp; g->sp->ucopy = val.ucopy;
	prArrayInsert(g, 3);
	
	scArray = (g->sp-2)->uo;

	if(scList)
		scList->slots[0].uo = scArray;
	
	g->sp = oldStack;
}

- (void)removeLastObject {
	VMGlobals *g = gMainVMGlobals;
	PyrSlot val, *oldStack;

	oldStack = g->sp;

	++g->sp; SetObject((PyrSlot *)g->sp, scArray);
	prArrayPop(g, 1);
	
	g->sp = oldStack;
}

- (void)removeObjectAtIndex:(unsigned)idx {
	VMGlobals *g = gMainVMGlobals;
	PyrSlot val, *oldStack;

	oldStack = g->sp;

	++g->sp; SetObject((PyrSlot *)g->sp, scArray);
	++g->sp; SetInt((PyrSlot *)g->sp, (int)idx);
	basicRemoveAt(g, 2);
	
	g->sp = oldStack;
}

@end