#import <CoreFoundation/CoreFoundation.h>
#import <Cocoa/Cocoa.h>
#import <QTKit/QTKit.h>
#import <objc/objc.h>
#import "CCBridge.h"
#import "CCPrimitives.h"
#import "CCMetadata.h"
#import "BetterInvocation.h"
#import "CCArray.h"
#import "next_mapping.h"
#import "ffi_utils.h"
#import "ffi.h"
#import "SCVirtualMachine.h"
#import "VMGlobals.h"
#import "MyDocument.h"
#import "GC.h"
#import "PyrPrimitive.h"
#import "PyrObject.h"
#import "PyrKernel.h"
#import "PyrKernelProto.h"
#import "PyrInterpreter.h"
#import "PyrSched.h"

#define NSLog(format, ...)

CCMetadata *metadata = nil;
NSMapTable *scObjcObjects = nil;
NSMutableArray *customClasses = nil;
extern bool compiledOK;
extern PyrSymbol *s_cctrampoline;
extern PyrClass *gClassList;

int finalizer(struct VMGlobals *g, struct PyrObject *obj);
int slotGetNSRect(PyrSlot* a, NSRect *r);
int slotGetPoint(PyrSlot* a, NSPoint *r);

@class NSPlaceholderString;
@class NSPlaceholderMutableString;
@class NSPlaceholderArray;
@class NSPlaceholderMutableArray;

/* to supress warnings */
@interface NSMethodSignature(signatureWithObjCTypes)
+ (NSMethodSignature *)signatureWithObjCTypes:(char*)types;
@end

int
objc_to_sc(const char *type, void *data, PyrSlot *slot, BOOL finalize) {
	VMGlobals *g = gMainVMGlobals;
	
	@try {
		type = objc_skip_type_qualifiers(type);
		
		switch(*type) {
			case _C_CHR:
				/*
				 * All chars are considered to be BOOL because of how rarely Cocoa returns a char (BOOL is typedefed to char)
				 * If a situation comes up where chars are necessary a special type code could be come up with to use in the method lookup file.
				 */
				if((BOOL)(*(char *)data)) SetTrue(slot);
				else SetFalse(slot);
				break;
				
			case _C_UCHR:
				SetInt(slot, (int)(*(unsigned char *)data));
				break;
				
			case _C_SEL:
			case _C_CHARPTR:
			case _C_ATOM:
			{
				PyrObject *scObj = instantiateObject(g->gc, findsym("CCString")->u.classobj, 1, false, false);
				PyrSlot *slots = scObj->slots;
				SetPtr(slots+0, [NSString stringWithUTF8String:(*(char **)data)]);
				SetObject(slot, scObj);
				
				//SetObjectOrNil(slot, (PyrObject *)newPyrString(g->gc, (*(char **)data), 0, true));
				break;
			}
				
			case _C_BOOL:
				if(*(bool*)data) SetTrue(slot);
				else SetFalse(slot);
				break;
				
			case _C_INT:
				SetInt(slot, (int)(*(int *)data));
				break;
				
			case _C_UINT:
				SetFloat(slot, (double)(*(unsigned int *)data));
				break;
				
			case _C_SHT:
				SetInt(slot, (int)(*(short *)data));
				break;
				
			case _C_USHT:
				SetInt(slot, (int)(*(unsigned short *)data));
				break;
				
			case _C_LNG:
				SetFloat(slot, (double)(*(long *)data));
				break;
				
			case _C_ULNG:
				SetFloat(slot, (double)(*(unsigned long *)data));
				break;
				
			case _C_LNG_LNG:
				SetFloat(slot, (double)(*(long long *)data));
				break;
				
			case _C_ULNG_LNG:
				SetFloat(slot, (double)(*(unsigned long long *)data));
				break;
				
			case _C_DBL:
				SetFloat(slot, (double)(*(double *)data));
				break;
				
			case _C_FLT:
				SetFloat(slot, (double)(*(float *)data));
				break;
				
			case _C_ID:
			{
				id obj = *(id *)data;
				
				/* In theory this is a no-op, in practice this gives us EOF 4.5 
				* support.
				*
				* EOF can return references to 'to-be-restored' objects, 
				* calling any method on them fully restores them, 'self' is
				* the safest method to call.
				*/
				//obj = [obj self];
				
				if(obj) {
					NSLog(@"creating supercollider object of class %s for object: %p", obj->isa->name, obj);
					if([obj isKindOfClass:[NSPlaceholderString class]] || 
						[obj isKindOfClass:[NSPlaceholderMutableString class]] ||
						[obj isKindOfClass:[NSPlaceholderArray class]] ||
						[obj isKindOfClass:[NSPlaceholderMutableArray class]]) {
						PyrSymbol *sym = findsym(obj->isa->name);
						PyrObject *scObject = instantiateObject(g->gc, sym->u.classobj, 2, false, false);
						PyrSlot *scObjectSlots = scObject->slots;
						SetPtr(scObjectSlots+0, obj);
						SetObject(slot, scObject);
					}
					else if([obj isKindOfClass:[NSNumber class]]) {
						const char *numType = [obj objCType];
						char *val = (char *)malloc(objc_sizeof_type(numType));

						NSLog(@"converting number of type %s to sc\n", numType);

						[obj getValue:val];
						objc_to_sc(numType, &(*val), slot, NO);
						free(val);
					}
					else if([obj isKindOfClass:[NSString class]]) {
						PyrString* string;
						string = (PyrString*)g->gc->New([obj length], 0, obj_char, true);
						string->classptr = findsym("CCString")->u.classobj;
						string->size = [obj length];
						memcpy(string->s, [obj UTF8String], [obj length]);
						SetObject(slot, string);
					}
					else {
						PyrObject *scObject;
						
						object_getInstanceVariable(obj, "scObject", (void**)&scObject);

						if(!scObject && (!(scObject = (PyrObject *)NSMapGet(scObjcObjects, obj)) || !finalize)) {
							PyrSymbol *sym = findsym(obj->isa->name);
							
							if(!sym) {
								post("Objective-C class does not exist in SuperCollider: %s\n", obj->isa->name);
								SetNil(slot);
								break;
							}
							
							scObject = instantiateObject(g->gc, sym->u.classobj, 3, true, false);
							PyrSlot *scObjectSlots = scObject->slots;
							SetPtr(scObjectSlots+0, obj);
							
							if(finalize) {
								PyrObject *finalizerObj = g->gc->NewFinalizer(finalizer, scObject, false);
								SetObject(scObjectSlots + 1, finalizerObj);
								g->gc->GCWrite(scObject, finalizerObj);
								
								NSMapInsert(scObjcObjects, obj, scObject);
							}
							
							object_setInstanceVariable(obj, "scObject", scObject);
						}
						else {
							NSLog(@"got trampoline object for objc object either through scObject instance var. or through the hash table: %p", scObject);
						}
						
						NSLog(@"returning trampoline object: %p", scObject);
						SetObject(slot, scObject);
					}
				}
				else
					SetNil(slot);
				
				break;
			}
				
			case _C_CLASS:
			{
				Class c = *(Class *)data;
				if(c) SetObject(slot, newPyrString(g->gc, c->name, 0, true));
				else SetNil(slot);
				break;
			}
			
			/* not tested */
			case _C_PTR:
			{
				PyrObject *scObj = instantiateObject(g->gc, findsym("Ref")->u.classobj, 1, false, false);
				PyrSlot *slots = scObj->slots;
				objc_to_sc(type+1, *(void **)data, slots+0, YES);
				SetObject(slot, scObj);
				break;
			}
				
			case _C_UNION_B:
				break;
			
			/* it is possible to support dynamic structure definitions a la pyobjc eventually... */
			case _C_STRUCT_B:
			{
				//NSLog(@"struct type: %s");
				
				if(strcmp(type, @encode(NSRect)) == 0) {
					NSLog(@"is a rect");
					PyrObject *scObj = instantiateObject(g->gc, s_rect->u.classobj, 4, false, false);
					PyrSlot *slots = scObj->slots;
					objc_to_sc(@encode(float), &((NSRect *)data)->origin.x, slots+0, nil);
					objc_to_sc(@encode(float), &((NSRect *)data)->origin.y, slots+1, nil);
					objc_to_sc(@encode(float), &((NSRect *)data)->size.width, slots+2, nil);
					objc_to_sc(@encode(float), &((NSRect *)data)->size.height, slots+3, nil);
					SetObject(slot, scObj);
				}
				else if(strcmp(type, @encode(NSPoint)) == 0) {
					NSLog(@"is a point");
					
					PyrObject *scObj = instantiateObject(g->gc, s_point->u.classobj, 2, false, false);
					PyrSlot *slots = scObj->slots;
					objc_to_sc(@encode(float), &((NSPoint *)data)->x, slots+0, nil);
					objc_to_sc(@encode(float), &((NSPoint *)data)->y, slots+1, nil);
					SetObject(slot, scObj);
				}
				else if(strcmp(type, @encode(NSRange)) == 0) {
					NSLog(@"is a range");
					
					PyrObject *scObj = instantiateObject(g->gc, findsym("Range")->u.classobj, 2, false, false);
					PyrSlot *slots = scObj->slots;
					objc_to_sc(@encode(int), &((NSRange *)data)->location, slots+0, nil);
					objc_to_sc(@encode(int), &((NSRange *)data)->length, slots+1, nil);
					SetObject(slot, scObj);				
				}
				else if(strcmp(type, @encode(NSSize)) == 0) {
					NSLog(@"is a size");
					
					PyrObject *scObj = instantiateObject(g->gc, findsym("Size")->u.classobj, 2, false, false);
					PyrSlot *slots = scObj->slots;
					objc_to_sc(@encode(float), &((NSSize *)data)->width, slots+0, nil);
					objc_to_sc(@encode(float), &((NSSize *)data)->height, slots+1, nil);
					SetObject(slot, scObj);
				}
				
				break;
			}
				
			case _C_ARY_B:
				break;
				
			case _C_VOID:
				break;
		}
	}
	@catch(NSException *e) {
		post([[NSString stringWithFormat:@"Exception while converting Obj-C type to SC: %@\n", [e reason]] UTF8String]);
	}
	
	//char str[256];
	//slotOneWord(slot, str); 
	//NSLog(@"objc_to_sc: %s", str);
	
	NSLog(@"done in objc_to_sc");
	
	return errNone;
}

int
sc_to_objc(const char *type, PyrSlot *slot, void *data) {
	NSLog(@"sc_to_objc: %s", type);

	@try {
		type = objc_skip_type_qualifiers(type);
		
		if(IsInt(slot)) {
			int val;
			slotIntVal(slot, &val);
			
			if(*type == _C_INT || *type == _C_UINT || *type == _C_LNG || *type == _C_ULNG ||  *type == _C_CHR) {
				memcpy(data, &val, sizeof(int));
				return errNone;
			}
			else if(*type == _C_ID) {
				(*(id *)data) = [NSNumber numberWithInt:val];
				return errNone;
			}
			
			post("wrong type, expected %c\n", *type);
			return errWrongType;
		}
		else if(IsFloat(slot)) {
			float fVal;
			int iVal;
			
			slotFloatVal(slot, &fVal);
			
			if(*type == _C_INT || *type == _C_UINT || *type == _C_LNG || *type == _C_ULNG ||  *type == _C_CHR) {
				iVal = (int)fVal;
				memcpy(data, &iVal, sizeof(int));
				return errNone;
			}
			else if(*type == _C_FLT || *type == _C_DBL) {		
				memcpy(data, &fVal, sizeof(float));
				return errNone;
			}
			else if(*type == _C_ID) {
				(*(id *)data) = [NSNumber numberWithFloat:fVal];
				return errNone;
			}

			post("wrong type, expected %c\n", *type);
			return errWrongType;
		}
		else if(IsFalse(slot) || IsTrue(slot)) {
			if(!(*type == _C_CHR)) {
				post("Wrong type. Objective-C is not expecting a BOOL.");
				return errWrongType;
			}
			
			BOOL val;
			if(IsTrue(slot)) val = YES;
			else val = NO;
			
			NSLog(@"converting BOOL, got: %d", val);
			memcpy(data, &val, sizeof(BOOL));
			
			return errNone;
		}
		else if(IsObj(slot) && isSubclassOf(slot->uo->classptr, s_cctrampoline->u.classobj)) {
			if(*type != _C_ID) return errWrongType;
			id val = (id) slot->uo->slots[0].uptr;
			
			NSLog(@"in trampoline converter %@", val);
			
			(*(id *)data) = val;
			
			return errNone;
		}
		/* add support for C style arrays... */
		else if(isKindOfSlot(slot, s_array->u.classobj)) {
			if(*type != _C_ID) return errWrongType;
			
			PyrObject *array = slot->uo;
			CCArray *newArray = [CCArray newWithSCArray:array];
			
			(*(id *)data) = newArray;
			
			return errNone;
		}
		else if(isKindOfSlot(slot, s_list->u.classobj)) {
			if(*type != _C_ID) return errWrongType;
			
			PyrObject *list = slot->uo;
			CCArray *newArray = [CCArray newWithSCList:list];
			
			NSLog(@"new ccarray list: %@\n", newArray);
			
			(*(id *)data) = newArray;
			
			return errNone;
		}
		else if(isKindOfSlot(slot, findsym("CCArray")->u.classobj)) {
		
		
		}
		else if (IsSym(slot) || isKindOfSlot(slot, s_string->u.classobj) || isKindOfSlot(slot, findsym("CCString")->u.classobj)) {
			char *str;

			if(IsSym(slot)) {
				str = slot->us->name;
			}
			else {
				PyrString *pyrString = slot->uos;
				str = (char *)malloc(pyrString->size + 1);
				memcpy(str, pyrString->s, pyrString->size);
				str[pyrString->size] = 0;
			}
		
			if(*type == _C_ID) {
				(*(id *)data) = [NSString stringWithUTF8String:str];
				if(!IsSym(slot)) free(str);
				NSLog(@"str in sc_to_objc: %@", (*(char **)data));
			}
			else if(*type == _C_CHARPTR) {
				(*(char **)data) = str;
			}
			else if(*type == _C_SEL) {
				(*(SEL *)data) = sel_registerName(str);
				if(!IsSym(slot)) free(str);
			}
			else return errWrongType;
			
			return errNone;
		}
		else if(isKindOfSlot(slot, s_rect->u.classobj)) {
			NSRect rect;
			
			if(*type != _C_STRUCT_B) return errWrongType;
			slotGetNSRect(slot, &rect);
			memcpy(data, &rect, sizeof(NSRect));
			return errNone;
		}
		else if(isKindOfSlot(slot, s_point->u.classobj)) {
			NSPoint point;
			
			if(*type != _C_STRUCT_B) return errWrongType;
			
			slotGetPoint(slot, &point);
			memcpy(data, &point, sizeof(NSPoint));
			return errNone;
		}
		else if(isKindOfSlot(slot, findsym("Size")->u.classobj)) {
			PyrSlot *slots = slot->uo->slots;
			NSSize size;
			int err;
			
			if(*type != _C_STRUCT_B) return errWrongType;
			
			err = slotFloatVal(slots+0, (float *)&size.width);
			if (err) return err;
			err = slotFloatVal(slots+1, (float *)&size.height);
			if (err) return err;
			
			memcpy(data, &size, sizeof(NSSize));
			return errNone;
		}
		else if(isKindOfSlot(slot, findsym("Range")->u.classobj)) {
			PyrSlot *slots = slot->uo->slots;
			NSRange range;
			int err;
			
			NSLog(@"in range handler");
			
			if(*type != _C_STRUCT_B) return errWrongType;
			
			err = slotIntVal(slots+0, (int *)&range.location);
			if (err) return err;
			err = slotIntVal(slots+1, (int *)&range.length);
			if (err) return err;
			
			memcpy(data, &range, sizeof(NSRange));
			return errNone;
		}
		else if(isKindOfSlot(slot, s_dictionary->u.classobj)) {
			post("A dictionary was sent to the Cocoa Bridge...which should never happen");
		}
		else if(isKindOfSlot(slot, findsym("CCDictWrapper")->u.classobj)) {
			if(*type != _C_ID) return errWrongType;
			
			unsigned int size = slot->uo->slots[0].uo->size;
			NSMutableDictionary *dict = [NSMutableDictionary dictionaryWithCapacity:size];
			
			NSLog(@"is a dict wrapper about to be pumped into an nsdictionary, size: %d", size);
			
			for(int i=0; i < size; i++) {
				PyrSlot key, value;
				id cKey, cValue;
				getIndexedSlot(slot->uo->slots[0].uo, &key, i);
				getIndexedSlot(slot->uo->slots[1].uo, &value, i);
				
				sc_to_objc(@encode(id), &key, &cKey);
				sc_to_objc(@encode(id), &value, &cValue);
				
				NSLog(@"%@:%@", cKey, cValue);
				
				[dict setObject:cValue forKey:cKey];
			}
			
			(*(id *)data) = dict;
			return errNone;
		}
	}
	@catch(NSException *e) {
		post([[NSString stringWithFormat:@"Exception while converting SC type to Obj-C: %@\n", [e reason]] UTF8String]);
	}
		
	NSLog(@"at the end of sc_to_objc");
	
	if(IsObj(slot)) NSLog(@"returning 0 in sc_to_objc. unhandled class: %s", slot->uo->classptr->name.us->name);
	
	(*(void **)data) = nil;
	
	NSLog(@"very end");
	return errWrongType;
}

SEL
CC_SCMethodNameToSelector(char *methodName) {
	char *newMethodName;
	NSMutableString *name = [NSMutableString stringWithUTF8String:methodName];
	[name replaceOccurrencesOfString:@"_" withString:@":" options:nil range:NSMakeRange(0, [name length])];
	
	NSLog(@"name: '%@', len: %d", name, [name length]);

	return NSSelectorFromString(name);
}

int identDictPut(struct VMGlobals *g, PyrObject *dict, PyrSlot *key, PyrSlot *value);
void Interpret(VMGlobals *g);

extern BOOL inCocoaCall;

int finalizer(struct VMGlobals *g, struct PyrObject *obj) {
	id val = (id) obj->slots[0].uptr;

	if(NSMapGet(scObjcObjects, val)) {
		NSMapRemove(scObjcObjects, val);
		[val release];
	}
}

static void
CC_Trampoline(ffi_cif* cif, void* resp, void** args, void* userdata) {
	id self = *(id *)args[0];
	SEL _cmd = *(SEL *)args[1];
	NSMethodSignature *methodSignature = (NSMethodSignature *)userdata;
	VMGlobals *g = gMainVMGlobals;
	PyrObject *scObject;
	PyrSlot tmpSCObject;
	
	NSLog(@"in CC_Trampoline, nargs: %d, sig: %@", cif->nargs, methodSignature);
	NSLog(@"_cmd: %s", _cmd);
	NSLog(@"self: %@", self);
	//NSLog(@"%s received message %s with signature %@", [self class]->name, _cmd, methodSignature);
	
	if(!CC_RespondsToSelector(self, _cmd)) {
		NSLog(@"returning early");
		return;
	}
	
	if(inCocoaCall) {
		post("A message was sent from SuperCollider to Obj-C then back to SuperCollider (SC->C->SC). This is not allowed; look into using .defer.\n");
		return;
	}
	else {
		NSLog(@"in cocoa call: %d", inCocoaCall);
		pthread_mutex_lock(&gLangMutex);
	}
	
	objc_to_sc(@encode(id), &self, &tmpSCObject, YES);
	scObject = tmpSCObject.uo;
	
	if(compiledOK && scObject) {
		NSMutableString *scSelector = [NSMutableString stringWithUTF8String:(const char *)_cmd];
		PyrSymbol *selector;
		int numArgs = 0;
		
		numArgs = [scSelector replaceOccurrencesOfString:@":" withString:@"_" options:nil range:NSMakeRange(0, [scSelector length])];
		selector = findsym([scSelector UTF8String]);
		
		if(selector) {
			PyrSlot result;
			VMGlobals *g = gMainVMGlobals;
			int i = 0;
			
			g->canCallOS = true;
			++g->sp;  SetObject(g->sp, scObject);
			
			for(i=2; i < [methodSignature numberOfArguments]; i++) {
				++g->sp;
				
				NSLog(@"size of %s: %d", [methodSignature getArgumentTypeAtIndex:i], objc_sizeof_type([methodSignature getArgumentTypeAtIndex:i]));;
				
				objc_to_sc([methodSignature getArgumentTypeAtIndex:i], args[i], g->sp, NO);
			}
			
			runInterpreter(g, selector, ([methodSignature numberOfArguments]-1));
			g->canCallOS = false;
			
			result.ucopy = g->result.ucopy;
			
			PyrSlot *slot = &g->result;
			const char *retType = [methodSignature methodReturnType];
			
			sc_to_objc(retType, slot, resp);
			
			NSLog(@"result is obj?: %d", IsObj(&g->result));
		}
	}
	
	pthread_mutex_unlock(&gLangMutex);
}

BOOL CC_RespondsToSelector(id self, SEL aSelector) {
	if(!aSelector) return NO;
	
	NSMutableString *scSelector = [NSMutableString stringWithUTF8String:(char *)aSelector];
	PyrClass *scClass;
	PyrMethod *meth;
	PyrSymbol *sym = NULL;
	int index, numArgs = 0;
	
	NSLog(@"in respondsToSelector for %s (%s)...", [self class]->name, aSelector);
	
	numArgs = [scSelector replaceOccurrencesOfString:@":" withString:@"_" options:nil range:NSMakeRange(0, [scSelector length])];
	
	if(sym = findsym([self class]->name))
		scClass = sym->u.classobj;
	else
		return NO;

	if(classFindDirectMethod(scClass, findsym([scSelector UTF8String]))) {
		NSLog(@"-respondsToSelector, method exists");
		return YES;
	}
	
	NSLog(@"-respondsToSelector, no such method");
	return NO;
}

char *
CC_TypesForSelector(SEL aSelector, Class klass, BOOL isClassMethod) {
	char *types;
	NSString *selector = NSStringFromSelector(aSelector);
	NSLog(@"in CC_TypesForSelector for %@", selector);
	
	/* first see if the method exists in klass (or its supers) */
	if(klass) {
		Method method;

		if(isClassMethod == YES)
			method = class_getClassMethod(klass, aSelector);
		else
			method = class_getInstanceMethod(klass, aSelector);
		
		if(method) return strdup(method->method_types);
	}

	if(types = [metadata informalProtocolMethodEncoding:aSelector]) return strdup(types);
	
	NSLog(@"in CC_TypesForSelector...2");
	
	/* if all else fails do what objective-c does and assume that the method takes only objects as arguments and returns an object */
	NSMutableString *signature = [NSMutableString stringWithString:@"@@:"];
	char c;
	while(c = *(((char *)aSelector)++)) {
		if(c == ':') [signature appendString:@"@"];
	}
	
	NSLog(@"sig for non-protocoled method: %@", signature);
	
	return strdup([signature UTF8String]);
}

NSMethodSignature* CC_MethodSignatureForSelector(SEL aSelector, Class klass, BOOL isClassMethod) {
	char *types = CC_TypesForSelector(aSelector, klass, isClassMethod);
	NSLog(@"in CC_MethodSignatureForSelector");
	
	/* should types be freed? does signatureWithObjCTypes: dup it? */
	
	if(types) {
		NSLog(@"in CC_MethodSignatureForSelector: %s, %s", [[NSMethodSignature signatureWithObjCTypes:types] methodReturnType], types);
		
		return [NSMethodSignature signatureWithObjCTypes:types];
	}
	else
		return nil;
}

Class
CC_CreateClass(const char *name, const char *superclassName) {
	struct objc_class * meta_class;
	struct objc_class * super_class;
	struct objc_class * new_class;
	struct objc_class * root_class;
	
	// Ensure that the superclass exists and that someone
	// hasn't already implemented a class with the same name
	//
	super_class = (struct objc_class *)objc_lookUpClass (superclassName);
	if (super_class == nil)
	{
		return nil;
	}
	
	if (objc_lookUpClass (name) != nil) 
	{
		NSLog(@"class %s already exists...not rebuilding", name);
		return nil;
	}
	
	// Find the root class
	//
	root_class = super_class;
	while( root_class->super_class != nil )
	{
		root_class = root_class->super_class;
	}
	
	// Allocate space for the class and its metaclass
	//
	new_class = (objc_class *)calloc( 2, sizeof(struct objc_class) );
	meta_class = &new_class[1];
	
	// setup class
	new_class->isa      = meta_class;
	new_class->info     = CLS_CLASS;
	meta_class->info    = CLS_META;
	
	// Create a copy of the class name.
	// For efficiency, we have the metaclass and the class itself 
	// to share this copy of the name, but this is not a requirement
	// imposed by the runtime.
	//
	new_class->name = (const char *)malloc (strlen (name) + 1);
	strcpy ((char*)new_class->name, name);
	meta_class->name = new_class->name;
	
	new_class->methodLists = nil;
	meta_class->methodLists = nil;
	
	// Connect the class definition to the class hierarchy:
	// Connect the class to the superclass.
	// Connect the metaclass to the metaclass of the superclass.
	// Connect the metaclass of the metaclass to the metaclass of  the root class.
	//
	new_class->super_class  = super_class;
	meta_class->super_class = super_class->isa;
	meta_class->isa         = (objc_class *)root_class->isa;
	
	// Set the sizes of the class and the metaclass.
	//
	new_class->instance_size = super_class->instance_size;
	meta_class->instance_size = meta_class->super_class->instance_size;
	
	// Finally, register the class with the runtime.
	//
	objc_addClass( new_class ); 
	return new_class;
}

struct objc_method_list *objc_allocMethodList(int numMethods)
{
	struct objc_method_list *mlist;
	
	if(numMethods <= 0) return NULL;
	
	mlist = (objc_method_list *)malloc(sizeof(struct objc_method_list) + ((numMethods+1) * sizeof(struct objc_method)));
	
	if (mlist == NULL)
		return NULL;
	
	mlist->method_count = 0;
	mlist->obsolete = NULL; 
	
	return mlist;
}

void objc_freeMethodList(struct objc_method_list **list)
{
	if (list)
	{
      if (list[0])
			free(list[0]);
		
      free(list);
	}
}

BOOL
automaticallyNotifiesObserversForKey(id self, SEL op, id key) {
	return NO;
}

BOOL
CC_CreateTrampoline(PyrClass *scClass) {
	struct objc_method_list *methodList, *classMethodList;
	Class klass;
	PyrSlot *scMethods = NULL;
	int numMethods = 0;
	
	NSLog(@"creating trampoline, class name: %s, super name: %s", scClass->name.us->name, scClass->superclass.us->u.classobj->name.us->name);
	
	klass = objc_lookUpClass(scClass->name.us->name);
	if(!klass) {
		NSLog(@"creating new class (can't reuse)");
		klass = CC_CreateClass(scClass->name.us->name, scClass->superclass.us->u.classobj->name.us->name);
		if(!klass) return nil;
	}
	
	if(scClass->methods.uo) {
		scMethods = scClass->methods.uo->slots;
		numMethods = scClass->methods.uo->size;
	}
	
	/* In order to support KVO we will need to notify observers ourselves (in CC_Trampoline). 
	 * This is because KVO uses isa swizzling to replace a class that is KVO'd with a class
	 * like NSKVONotifying_Foo, and once this class is built it isn't updated. This breaks
	 * adding a method to an obj-c subclass in sc and recompiling, because NSKVONotifying_Foo
	 * doesn't get the updated method. Returning NO from automaticallyNotifiesObserversForKey:
	 * disables the isa swizzling.
	 */
	
	if(!klass->isa->methodLists) {
		classMethodList = objc_allocMethodList(1);
		struct objc_method tmpKVOFix;
		
		tmpKVOFix.method_name = @selector(automaticallyNotifiesObserversForKey:);
		tmpKVOFix.method_types = CC_TypesForSelector(tmpKVOFix.method_name, [NSObject class], YES);
		tmpKVOFix.method_imp  = (IMP)automaticallyNotifiesObserversForKey;
		
		classMethodList->method_list[0] = tmpKVOFix;
		classMethodList->method_count = 1;
		class_addMethods(klass->isa, classMethodList);
	}

	methodList = objc_allocMethodList(numMethods);
	
	for(int i=0; i < numMethods; i++) {
		PyrMethod *scMethod = scMethods[i].uom;
		struct objc_method method;
		
		NSLog(@"length of method named: %s, %d", scMethod->name.us->name, scMethod->name.us->length);
		
		SEL selector = CC_SCMethodNameToSelector(scMethod->name.us->name);
		
		char *objcTypes = CC_TypesForSelector(selector, [klass superclass], NO);
		NSMethodSignature *methodSignature = [NSMethodSignature signatureWithObjCTypes:objcTypes];
		int numArgs = [methodSignature numberOfArguments];
		ffi_type **argTypes, *retType;
		ffi_cif *cif;
		ffi_closure *closure;
		
		NSLog(@"looking at %s, %s, numArgs: %d, (issetter?: %d)", scMethod->name.us->name, selector, numArgs, (scMethod->name.us->flags & sym_Setter));
		
		argTypes = (ffi_type **)malloc(sizeof(ffi_type *) * numArgs);
		
		for(int j=0; j < numArgs; j++) {
			argTypes[j] = signature_to_ffi_type((char *)[methodSignature getArgumentTypeAtIndex:j]);
		}
		
		retType = signature_to_ffi_type((char *)[methodSignature methodReturnType]);
		
		cif = (ffi_cif *)malloc(sizeof(ffi_cif));
		if (cif == NULL) {
			NSLog(@"Can't allocate memory");
			continue;
		}
		
		if (ffi_prep_cif(cif, FFI_DEFAULT_ABI, numArgs, retType, argTypes) != FFI_OK) {
			NSLog(@"Can't prepare CIF");
			continue;
		}
		
		closure = (ffi_closure *)malloc(sizeof(ffi_closure));
		if (closure == NULL) {
			NSLog(@"Can't allocate memory");
			continue;
		}
		
		if (ffi_prep_closure(closure, cif, CC_Trampoline, methodSignature) != FFI_OK) {
			NSLog(@"Can't prepare closure");
			continue;
		}

		method.method_name = selector;
		method.method_types = objcTypes;
		method.method_imp  = (IMP)closure;
		
		methodList->method_list[methodList->method_count] = method;
		methodList->method_count++;
	}
	
	//NSLog(@"objc method list count: %d, sc method list count: %d", methodList->method_count, numMethods);

	if(klass->methodLists) {
		struct objc_method_list *oldMethodList = (struct objc_method_list *)klass->methodLists;
		class_removeMethods(klass, oldMethodList);
		
		for (int j = 0; j < oldMethodList->method_count; j++) {
			objc_method m = oldMethodList->method_list[j];
			free(m.method_types);
		}
		
		free(oldMethodList);
	}
	
	if(methodList && (methodList->method_count > 0))
		class_addMethods(klass, methodList);
	
	struct objc_ivar ivar;
	ivar.ivar_name = "scObject";
	ivar.ivar_type = @encode(PyrObject *);
	ivar.ivar_offset = klass->super_class->instance_size;
	
	struct objc_ivar_list *instanceVars = (objc_ivar_list *)malloc(sizeof(struct objc_ivar_list) + (2 * sizeof(struct objc_ivar)));
	instanceVars->ivar_count = 1;
	instanceVars->ivar_list[0] = ivar;
	
	klass->ivars = instanceVars;
	klass->instance_size += sizeof(PyrObject *);
	
	return YES;
}

void
CC_FindCustomClasses(PyrClass *klass) {
	PyrObject *subclasses;
	Class customClass;
	
	if(objc_lookUpClass(klass->name.us->name) == nil) {
		[customClasses addObject:[NSString stringWithUTF8String:klass->name.us->name]];
	}
	
	subclasses = klass->subclasses.uo;
	
	if(subclasses) {
		for(int i = 0; i < (int)subclasses->size; i++) {
			CC_FindCustomClasses(subclasses->slots[i].uoc);
		}
	}
	
	return;
}

/* Traverse the tree in order from NSObject (super classes must be constructed before subclasses) */
BOOL
CC_CreateTrampolines() {
	for(int i=0; i < [customClasses count]; i++) {
		NSString *klass = [customClasses objectAtIndex:i];
		PyrSymbol *sym = findsym([klass UTF8String]);
		
		if(!sym || !sym->u.classobj) {
			[customClasses removeObjectAtIndex:i];
			i--;
		}
		else CC_CreateTrampoline(sym->u.classobj);
	}

	return YES;
}

#if 0
// this is not currently used, but may be eventually when support for dynamic struct. definitions is added
BOOL
CC_SetupStructureDefs() {
	NSMutableDictionary *defs = [[NSMutableDictionary alloc] init];
	NSDictionary *structure;
	
	structure = [NSDictionary dictionaryWithObjectsAndKeys:	@"struct NSSize(width, height)", @"doc",
		[NSArray arrayWithObjects:@"width", @"height", nil], @"elements",
		nil];
	[defs setObject:structure forKey:[NSString stringWithUTF8String:@encode(NSSize)]];
	
	
	structure = [NSDictionary dictionaryWithObjectsAndKeys:	@"struct NSRange(location, length)", @"doc",
		[NSArray arrayWithObjects:@"location", @"length", nil], @"elements",
		nil];
	[defs setObject:structure forKey:[NSString stringWithUTF8String:@encode(NSRange)]];
	
	
	structure = [NSDictionary dictionaryWithObjectsAndKeys:	@"struct NSAffineTransformStruct(m11, m12, m21, m22, tX, tY)", @"doc",
		[NSArray arrayWithObjects:@"m11", @"m12", @"m21", @"m22", @"tX", @"tY", nil], @"elements",
		nil];
	[defs setObject:structure forKey:[NSString stringWithUTF8String:@encode(NSAffineTransformStruct)]];
	
	
	structure = [NSDictionary dictionaryWithObjectsAndKeys:	@"struct QTTime(timeValue, timeScale, flags)", @"doc",
		[NSArray arrayWithObjects:@"timeValue", @"timeScale", @"flags", nil], @"elements",
		nil];
	[defs setObject:structure forKey:[NSString stringWithUTF8String:@encode(QTTime)]];	
	
	
	structure = [NSDictionary dictionaryWithObjectsAndKeys:	@"struct QTTimeRange(time, duration)", @"doc",
		[NSArray arrayWithObjects:@"time", @"duration", nil], @"elements",
		nil];
	[defs setObject:structure forKey:[NSString stringWithUTF8String:@encode(QTTimeRange)]];
	
	cocoaStructures = defs;
	
	return YES;
}
#endif

BOOL
CC_Setup() {
	NSEnumerator *enumerator;
	NSArray *mdPaths;
	NSString *path;
	//double startTime = elapsedTime();
	
	if(!compiledOK || !findsym("NSObject")) return NO;
	
	/* CC_Setup() is called everytime you compile. Release scObjcObjects upon compilation */
	if(scObjcObjects) NSResetMapTable(scObjcObjects);
	else {
		scObjcObjects = NSCreateMapTable(	NSNonOwnedPointerMapKeyCallBacks,
														NSNonOwnedPointerMapValueCallBacks,
														100);
	}
	
	/* Only initialize the metadata once. The program will need to be restarted if changes are made. */
	if(!metadata) {	
		metadata = [[CCMetadata alloc] init];
		mdPaths = [[NSBundle mainBundle] pathsForResourcesOfType:@"xml" inDirectory:@"CocoaCollider Metadata"];
		enumerator = [mdPaths objectEnumerator];
		while(path = [enumerator nextObject]) {
			NSLog(@"parsing: %@", path);
			[metadata parse:path];
		}
	}
	
	if(!customClasses) customClasses = [[NSMutableArray arrayWithCapacity:10] retain];
	CC_FindCustomClasses(findsym("NSObject")->u.classobj);
	
	CC_CreateTrampolines();
	
	//post("CocoaCollider initialized in %.2f seconds\n", elapsedTime()-startTime);
}