#import "CCMetadata.h"
#import <Cocoa/Cocoa.h>
#import <dlfcn.h>

// there is an issue getting this into C++
#define RTLD_DEFAULT    ((void *) -2)   /* Use default search algorithm. */

#define NSLog(...)

@implementation CCMetadata

/* need to add a custom call back to free constant type strings */

- (id)init {
	if(self = [super init]) {
		constants = NSCreateMapTable(	NSObjectMapKeyCallBacks,
												NSNonOwnedPointerMapValueCallBacks,
												100);
		informalProtocols = NSCreateMapTable(	NSNonOwnedPointerMapKeyCallBacks,
															NSOwnedPointerMapValueCallBacks,
															100);
	}
	
	return self;
}

- (BOOL)parse:(NSString *)path {
	NSError *error;
	NSXMLDocument *xmlDoc = [[NSXMLDocument alloc] initWithContentsOfURL:[NSURL fileURLWithPath:path] options:nil error:&error];
	NSXMLElement *root;

	if(!xmlDoc) {
		NSLog(@"Error while parsing %@: %@", path, error);
		return NO;
	}
	
	root = [xmlDoc rootElement];
	if(![[[root attributeForName:@"version"] stringValue] isEqualToString:@"0.9"]) {
		NSLog(@"%@ was generated with version %@ of the generator, CocoaCollider only handles version 0.9.", 
				[path lastPathComponent], [root attributeForName:@"version"]);
		return NO;
	}
	
	for(int i = 0; i < [root childCount]; i++) {
		NSXMLElement *node = (NSXMLElement *)[root childAtIndex:i];
		NSString *mdType = [node name];
		
		if([mdType isEqualToString:@"constant"] || [mdType isEqualToString:@"enum"]) {
			struct constant *mConstant;
			
			if(![[node attributeForName:@"name"] stringValue]) continue;
			
			mConstant = (struct constant *)malloc(sizeof(struct constant));
			
			if([mdType isEqualToString:@"constant"]) {
				void *data;
				
				if(![[node attributeForName:@"type"] stringValue]) {
					free(mConstant);
					continue;
				}
				
				data = dlsym(RTLD_DEFAULT, [[[node attributeForName:@"name"] stringValue] UTF8String]);
				
				if(!data) {
					free(mConstant);
					continue;
				}
				
				mConstant->type = strdup([[[node attributeForName:@"type"] stringValue] UTF8String]); 
				mConstant->data = data;
								
				if(strcmp(mConstant->type, @encode(id)) == 0)
					NSLog(@"constant: %@,%@", [[node attributeForName:@"name"] stringValue], *((id *)mConstant->data));
			}
			else if([mdType isEqualToString:@"enum"]) {
				NSString *value = [[node attributeForName:@"value"] stringValue];
				
				if(!value) {
					free(mConstant);
					continue;
				}
				
				if(strchr([value UTF8String], '.')) {
					double _dVal = [value doubleValue];
					void *dVal = malloc(sizeof(double));
					memcpy(dVal, &_dVal, sizeof(double));
					mConstant->type = @encode(double);
					mConstant->data = dVal;
					NSLog(@"enum: %@,%s,%f", [[node attributeForName:@"name"] stringValue], mConstant->type, _dVal);
				}
				else {
					int _iVal = [value intValue];
					void *iVal = malloc(sizeof(int));
					memcpy(iVal, &_iVal, sizeof(int));
					mConstant->type = @encode(int);
					mConstant->data = iVal;
					NSLog(@"enum: %@,%s,%d", [[node attributeForName:@"name"] stringValue], mConstant->type, _iVal);
				}
			}

			NSMapInsert(constants, [[node attributeForName:@"name"] stringValue], mConstant);
		}
		else if([mdType isEqualToString:@"informal_protocol"]) {
			NSXMLElement *ipRoot = node;
			
			for(int j=0; j < [ipRoot childCount]; j++) {
				NSXMLElement *ipNode = (NSXMLElement *)[ipRoot childAtIndex:j];
				if(![[ipNode attributeForName:@"selector"] stringValue] || ![[ipNode attributeForName:@"encoding"] stringValue]) continue;
				NSLog(@"informal_protocol: %@, %@", [[ipNode attributeForName:@"selector"] stringValue], [ipNode attributeForName:@"encoding"]);
				NSMapInsert(informalProtocols, NSSelectorFromString([[ipNode attributeForName:@"selector"] stringValue]), strdup([[[ipNode attributeForName:@"encoding"] stringValue] UTF8String]));
			}
		}
	
	}
	
	[xmlDoc release];

	return YES;
}

- (void *)constantForName:(NSString *)name type:(char **)type {
	struct constant *mConstant;
	mConstant = (struct constant *)NSMapGet(constants, name);
	if(mConstant) {
		if(type) *type = mConstant->type;
		return mConstant->data;
	}
	else return nil;
}

- (char *)informalProtocolMethodEncoding:(SEL)sel {
	return (char *)NSMapGet(informalProtocols, sel);
}

- (void)dealloc {
	NSFreeMapTable(constants);
	NSFreeMapTable(informalProtocols);	
}

@end
