#import <Cocoa/Cocoa.h>
#import <objc/objc.h>
#import "CCBridge.h"
#import "CCPrimitives.h"
#import "CCMetadata.h"
#import "BetterInvocation.h"
#import "next_mapping.h"
#import "SCVirtualMachine.h"
#import "VMGlobals.h"
#import "MyDocument.h"
#import "GC.h"
#import "PyrPrimitive.h"
#import "PyrObject.h"
#import "PyrKernel.h"
#import "PyrInterpreter.h"
#import "ffi.h"

#define NSLog(format, ...)

BOOL invocationReturnToSC(struct VMGlobals *g, NSInvocation *invocation, PyrSlot *ret);

BOOL inCocoaCall;
PyrSymbol *s_cctrampoline;
NSMutableArray *nsobjects;
extern bool compiledOK;
extern NSMapTable *scObjcObjects;
extern CCMetadata *metadata;

@implementation NSString(stringWithPyrSlot)
+ (NSString *)stringWithPyrSlot:(PyrSlot *)slot {	 
	 if(IsSym(slot) && slot->us) {
		return [[self class] stringWithCString:slot->us->name length:slot->us->length];
	 }
	 else if(IsObj(slot) && slot->uo) {
		return [[self class] stringWithCString:slot->uos->s length:slot->uos->size];
	 }
	 
	 return nil;
}
@end

@implementation SCVirtualMachine(deferedCallBack)
+ (void)makeDeferedCallBack:(int)callBackNum invocation:(NSInvocation *)invocation {
	VMGlobals *g = gMainVMGlobals;
	PyrSymbol *selector = findsym("deferedCallBack");
	PyrSlot *ret = (PyrSlot *)malloc(sizeof(PyrSlot));
	
	if(!selector) {
		post("deferedCallBack is not defined");
		return;
	}
	
	NSLog(@"in makeDeferedCallBack");
	
	@try {
		[invocation invoke];
	}
	@catch (NSException *e) {
		post([[NSString stringWithFormat:@"Exception: %@\n", [e reason]] UTF8String]);
	}
	
	NSLog(@"made invocation");

	pthread_mutex_lock(&gLangMutex);
	
	g->canCallOS = true;
	++g->sp; SetObject(g->sp, findsym("CCTrampoline")->u.classobj);
	++g->sp; SetInt(g->sp, callBackNum);
	++g->sp; invocationReturnToSC(g, invocation, g->sp);

	runInterpreter(g, selector, 3);
	g->canCallOS = false;
	
	pthread_mutex_unlock (&gLangMutex);
	
	[invocation release];
}
@end

void scMethodToObjC(NSMutableString *methodName) {
	[methodName replaceOccurrencesOfString:@"_" withString:@":" options:nil range:NSMakeRange(0, [methodName length])];
}

int prObjcRespondsTo(struct VMGlobals *g, int numArgsPushed);
int prObjcRespondsTo(struct VMGlobals *g, int numArgsPushed) {
	PyrSlot *a, *b;
	PyrClass *classobj;
	NSMutableString *method;
	SEL selector;
	
	a = g->sp - 1;
	b = g->sp;
	
	classobj = classOfSlot(a);
	
	if (b->utag == tagSym) {
		method = [NSMutableString stringWithPyrSlot:b];
		scMethodToObjC(method);
		selector = NSSelectorFromString(method);
		
		if(isKindOfSlot(a, class_class)) {
			if(!class_getClassMethod(objc_getClass(classobj->name.us->name+5), selector))
				a->ucopy = o_false.ucopy;
			else
				a->ucopy = o_true.ucopy;
		}
		else {
			if(!class_getInstanceMethod(objc_getClass(classobj->name.us->name), selector))
				a->ucopy = o_false.ucopy;
			else
				a->ucopy = o_true.ucopy;		
		}
	}
	else return errWrongType;
		
	return errNone;
}

int prObjcInvoke(struct VMGlobals *g, int numArgsPushed);
int prObjcInvoke(struct VMGlobals *g, int numArgsPushed) {
   PyrSlot *ret = g->sp - 6;
	PyrSlot *argClassName = g->sp - 5;
	PyrSlot *argIsClassMethod = g->sp - 4;
	PyrSlot *argMethod = g->sp - 3;
	PyrSlot *argArgs = g->sp - 2;
	PyrSlot *argObjcInstance = g->sp - 1;
	PyrSlot *argDeferNum = g->sp;
	
	NSLog(@"prObjcInvoke, %d, %d, %d, %d", isKindOfSlot(argClassName, class_string), 
														(IsFalse(argIsClassMethod) || IsTrue(argIsClassMethod)),
														isKindOfSlot(argMethod, class_string),
														isKindOfSlot(argArgs, class_array));
	
	if(!g->canCallOS) return errCantCallOS;
	
	if(!isKindOfSlot(argClassName, class_string) ||
		!(IsFalse(argIsClassMethod) || IsTrue(argIsClassMethod)) ||
		!isKindOfSlot(argMethod, class_string) ||
		!isKindOfSlot(argArgs, class_array)) return errFailed;
	
	NSLog(@"everything is correctly typed");
	
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	NSString *className = [NSString stringWithPyrSlot:argClassName];
	BOOL isClassMethod = IsTrue(argIsClassMethod);
	NSMutableString *method = (NSMutableString *)[NSMutableString stringWithPyrSlot:argMethod];
	id objcInstance;
	int numberOfArgs = argArgs->uo->size;
	int deferNum = -1;
	
	if(IsNil(argDeferNum)) deferNum = -1;
	else deferNum = argDeferNum->ui;
	
	if(IsPtr(argObjcInstance)) {
		objcInstance = (id)argObjcInstance->uop;
		//NSLog(@"objcInstance is a pointer, %@", objcInstance);
	}
	else
		sc_to_objc(@encode(id), argObjcInstance, &objcInstance);
	
	scMethodToObjC(method);
	
	NSLog(@"className: %@, isClassMethod: %d, method: %@, numberOfArgs: %d, deferNum: %d", className, isClassMethod, method, numberOfArgs, deferNum);
	NSLog(@"here -1, %p", objcInstance);
	if(objcInstance) NSLog(@"objInstance: %@", [objcInstance methodSignatureForSelector:NSSelectorFromString(method)]);
	NSLog(@"here 0");
	
	Class objcClass = NSClassFromString(className);
	SEL selector = NSSelectorFromString(method);
	NSMethodSignature *signature;
	
	NSLog(@"here1");
	
	if(isClassMethod) signature = [objcClass methodSignatureForSelector:selector];
	else signature = [objcInstance methodSignatureForSelector:selector];
	if(!signature) goto err;
	
	NSLog(@"here2");
	
	//DummyAutoreleasePool *pool = [[DummyAutoreleasePool alloc] init];
	BetterInvocation *invocation = (BetterInvocation *)[BetterInvocation invocationWithMethodSignature:signature];
	//[pool release];
	
	NSLog(@"here3, %s", selector);
	NSLog(@"signature: %@", signature);
	[invocation setSelector:selector];
	NSLog(@"here4");
	NSLog(@"built invocation: %@", invocation);
	
	for(int i=0; i < numberOfArgs; i++) {
		const char *argType = [signature getArgumentTypeAtIndex:i+2];
		void *argBuffer = malloc(objc_sizeof_type(argType));
		PyrSlot *arg = argArgs->uo->slots+i;
		
		sc_to_objc(argType, arg, argBuffer);
		[invocation setArgument:argBuffer atIndex:i+2];
	}
	
	NSLog(@"setting in cocoa call, %d", isClassMethod);
	BOOL alreadyInCocoaCall = inCocoaCall;

	if(isClassMethod) {
		[invocation setTarget:objcClass];
	}
	else {
		if([className isEqualToString:[NSString stringWithUTF8String:objcInstance->isa->name]])
			[invocation setTarget:objcInstance];
		else {
			[invocation setTarget:objcInstance];
			[invocation setSendsToSuper:YES];
		}
	}	
	
	if(deferNum >= 0) {
		SEL sel = @selector(makeDeferedCallBack:invocation:);
		NSMethodSignature *sig = [SCVirtualMachine methodSignatureForSelector:sel];
		NSInvocation *deferedInvocation = [NSInvocation invocationWithMethodSignature:sig];

		[deferedInvocation setTarget:[SCVirtualMachine class]];
		[deferedInvocation setSelector:sel];
		[deferedInvocation setArgument:&deferNum atIndex:2];
		[deferedInvocation setArgument:&invocation atIndex:3];
		
		[invocation retainArguments];
		[invocation retain];
		
		[[SCVirtualMachine sharedInstance] defer:deferedInvocation];
		SetNil(ret);
		goto done;
	}
	
	inCocoaCall = YES;
	//pool = [[DummyAutoreleasePool alloc] init];
	@try {
		[invocation invoke];
	}
	@catch(NSException *e) {
		post([[NSString stringWithFormat:@"Exception: %@\n", [e reason]] UTF8String]);
	}
	//[pool release];
	if(alreadyInCocoaCall == NO) inCocoaCall = NO;
	
	NSLog(@"done with call");
	
	invocationReturnToSC(g, invocation, ret);
	
	done:
		[pool release];
		return errNone;
		
	err:
		[pool release];
		return errFailed;
}

BOOL
invocationReturnToSC(struct VMGlobals *g, NSInvocation *invocation, PyrSlot *ret) {
	NSMethodSignature *signature = [invocation methodSignature];
	SEL selector = [invocation selector];
	const char *retType = [signature methodReturnType];
	
	if([signature methodReturnLength] > 0) {
		void *retBuffer = malloc([signature methodReturnLength]);
		BOOL finalize = YES;
		
		[invocation getReturnValue:retBuffer];
		
		if(*retType == _C_ID) {
			id objcObject = *(id *)retBuffer;

			if((!NSMapGet(scObjcObjects, objcObject))
					&& selector != @selector(alloc)
					&& selector != @selector(allocWithZone:)
					&& selector != @selector(new)
					&& selector != @selector(copy)
					&& selector != @selector(mutableCopy)) {
				[objcObject retain];
			}
			
			if(selector == @selector(alloc) || selector == @selector(allocWithZone:)) finalize = NO;
		}

		objc_to_sc(retType, retBuffer, ret, finalize);
	}
	else {
		SetNil(ret);
	}
	
	return YES;
}

int prObjcGetConstant(struct VMGlobals *g, int numArgsPushed);
int prObjcGetConstant(struct VMGlobals *g, int numArgsPushed) {
	PyrSlot *ret = g->sp - 1;
   PyrSlot *_name = g->sp;
	NSString *name = [NSString stringWithPyrSlot:_name];
	char *type = nil;
	void *constant = nil;
	
	if(!name) return errFailed;
	
	constant = [metadata constantForName:name type:&type];
	if(!constant || !type) {
		SetNil(ret);
		return errNone;
	}
	
	if(objc_to_sc(type, constant, ret, NO) != errNone) return errFailed;
	
	return errNone;
}

void initCocoaBridgePrimitives()
{
	int base = nextPrimitiveIndex(), index = 0;
	s_cctrampoline = getsym("CCTrampoline");
	definePrimitive(base, index++, "_ObjC_Invoke", prObjcInvoke, 7, 0);
	definePrimitive(base, index++, "_ObjC_RespondsTo", prObjcRespondsTo, 2, 0);
	definePrimitive(base, index++, "_ObjC_GetConstant", prObjcGetConstant, 2, 0);
}
