//
//  CCArrayTest.m
//  xSC3lang
//
//  Created by Ryan Brown on 12/16/06.
//  Copyright 2006 __MyCompanyName__. All rights reserved.
//

#import "CCArrayTest.h"
#import "CCTestUtils.h"

#import "SCVirtualMachine.h"
#import "VMGlobals.h"
#import "MyDocument.h"
#import "GC.h"
#import "PyrPrimitive.h"
#import "PyrObject.h"
#import "PyrKernel.h"
#import "PyrKernelProto.h"
#import "PyrInterpreter.h"

@implementation CCArrayTest
- (void) setUp
{
	VMGlobals *g = gMainVMGlobals;
	PyrObject *pyrArray;

	testAgainst = [[NSArray arrayWithObjects:@"zero", @"one", @"two", @"three", nil] retain];
	
	[CCTestUtils execute:@"a = List['zero', 'one', 'two', 'three']"];
	pyrArray = ((PyrInterpreter *)g->process->interpreter.uo)->a.uo;
	
	array = [CCArray newWithSCList:pyrArray];
	STAssertNotNil(array, @"returned CCArray is nil");
	
	[array retain];
}

- (void)testCount {
	STAssertEquals([array count], 4, @"CCArray created with 4 elements, but has %d", [array count]);
}

- (void)testObjectAtIndex {
	STAssertEqualObjects([array objectAtIndex:2], @"two", @"element 2 is '%@', should be 'two'", [array objectAtIndex:2]);
}

- (void)testReplaceObjectAtIndex {
	[array replaceObjectAtIndex:2 withObject:@"replaced_two"];
	STAssertEqualObjects([array objectAtIndex:2], @"replaced_two", @"object at index 2 is '%@', not 'replaced_two'", [array objectAtIndex:2]);
}

- (void)testEnumerator {
	NSEnumerator *enumerator = [array objectEnumerator];
	id obj;
	int i = 0;
	
	while(obj = [enumerator nextObject]) {
		STAssertEqualObjects(obj, [testAgainst objectAtIndex:i], @"element %d is '%@', not '%@'", i, obj, [testAgainst objectAtIndex:i]);
		i++;
	}
}

- (void)testGetObjects {
	NSRange range;
	id *buf = (id *)calloc(4, sizeof(id));

	range.location = 0;
	range.length = 4;
	
	[array getObjects:buf inRange:range];
	
	for(int i=0; i < 4; i++) STAssertEqualObjects(buf[i], [testAgainst objectAtIndex:i], @"element %d is '%@', not '%@'", i, *buf, [testAgainst objectAtIndex:i]);
}

- (void)testAddObject {
	[array addObject:@"four"];
	STAssertEqualObjects([array objectAtIndex:4], @"four", @"object at index four is '%@', not 'four'", [array objectAtIndex:4]);
}

- (void)testInsertObject {
	[array insertObject:@"zero_inserted" atIndex:0];
	STAssertEqualObjects([array objectAtIndex:0], @"zero_inserted", @"index 0 is not 'zero_inserted', is '%@'", [array objectAtIndex:0]);
	
	[array insertObject:@"five_inserted" atIndex:5];
	STAssertEqualObjects([array objectAtIndex:5], @"five_inserted", @"index 5 is not 'five_inserted', is '%@'", [array objectAtIndex:5]);	
}

- (void)testRemoveLastObject {
	[array removeLastObject];
	STAssertEquals([array count], 3, @"count should be 4 after remove, is %d", [array count]);
}

- (void)testRemoveObjectAtIndex {
	[array removeObjectAtIndex:1];
	STAssertEqualObjects([array objectAtIndex:1], @"two", @"index 1 should be 'two' after remove, is '%@'", [array objectAtIndex:1]);
}

- (void) tearDown
{
	[CCTestUtils execute:@"a = nil"];
	[array release];
}

@end
