#import <Cocoa/Cocoa.h>
#import <objc/objc-runtime.h>
#import "ffi_utils.h"
#import "next_mapping.h"
#import "ffi.h"

#define NSLog(format, ...)

size_t
num_struct_fields(const char* argtype)
{
	size_t res = 0;
	
	if (*argtype != _C_STRUCT_B) return -1;
	while (*argtype != _C_STRUCT_E && *argtype != '=') argtype++;
	if (*argtype == _C_STRUCT_E) return 0;
	
	argtype++;
	while (*argtype != _C_STRUCT_E) {
		argtype = objc_skip_typespec(argtype);
		if (argtype == NULL) return -1;
		res++;
	}
	return res;
}

void
free_type(ffi_type *obj)
{
	free(obj->elements);
	free(obj);
}

ffi_type* 
struct_to_ffi_type(const char* argtype)
{
	ffi_type *type;
	struct objc_struct_layout layout;
	int position, align;
	const char *curtype;
	size_t field_count;
	
	type = (ffi_type *)malloc(sizeof(ffi_type));
	if(type == NULL) {
		NSLog(@"no memory");
		return NULL;
	}
	
	field_count = num_struct_fields(argtype);
	if(field_count == -1) {
		NSLog(@"Cannot determine layout of %s", argtype);
		return NULL;
	}
	
	type->size = objc_sizeof_type(argtype);
	type->alignment = objc_alignof_type(argtype);
	type->type = FFI_TYPE_STRUCT;
	type->elements = (ffi_type **)malloc((1+field_count) * sizeof(ffi_type*));
	if (type->elements == NULL) {
		free(type);
		NSLog(@"no memory");
		return NULL;
	}
	
	field_count = 0;
	curtype = argtype+1;
	while (*curtype != _C_STRUCT_E && *curtype != '=') curtype++;
	if (*curtype == '=') {
		curtype ++;
		while (*curtype != _C_STRUCT_E) {
			type->elements[field_count] =  signature_to_ffi_type(curtype);
			if (type->elements[field_count] == NULL) {
				free(type->elements);
				return NULL;
			}
			field_count++;
			curtype = objc_skip_typespec(curtype);
			if (curtype == NULL) {
				free(type->elements);
				return NULL;
			}
		}
	}
	type->elements[field_count] = NULL;
	
	return type;
}

ffi_type *
signature_to_ffi_type(const char *argtype)
{
	switch (*argtype) {
		case _C_VOID: return &ffi_type_void;
		case _C_ID: return &ffi_type_pointer;
		case _C_CLASS: return &ffi_type_pointer;
		case _C_SEL: return &ffi_type_pointer;
		case _C_CHR: return &ffi_type_schar;
#ifdef _C_BOOL
		case _C_BOOL: return &ffi_type_sint;
#endif	
		case _C_UCHR: return &ffi_type_uchar;
		case _C_SHT: return &ffi_type_sshort;
		case _C_USHT: return &ffi_type_ushort;
		case _C_INT: return &ffi_type_sint;
		case _C_UINT: return &ffi_type_uint;
			
			/* The next two defintions are incorrect, but the correct definitions
			* don't work (e.g. give testsuite failures). We should be fine
			* as long as sizeof(long) == sizeof(int)
			*/
		case _C_LNG: return &ffi_type_sint;  /* ffi_type_slong */
		case _C_ULNG: return &ffi_type_uint;  /* ffi_type_ulong */
		case _C_LNG_LNG: return &ffi_type_sint64;
		case _C_ULNG_LNG: return &ffi_type_uint64;
		case _C_FLT: return &ffi_type_float;
		case _C_DBL: return &ffi_type_double;
		case _C_CHARPTR: return &ffi_type_pointer;
		case _C_PTR: return &ffi_type_pointer;
			
			/*
			 case _C_ARY_B: 
				 return array_to_ffi_type(argtype);
				 */
			
		case _C_IN: case _C_OUT: case _C_INOUT: case _C_CONST:
			return signature_to_ffi_type(argtype+1);
		case _C_STRUCT_B: 
			return struct_to_ffi_type(argtype);
		default:
			NSLog(@"signature_to_ffi_type: Type %c not supported", *argtype);
			return NULL;
	}
}